==> comedi-0.5.4.tgz: comedi-0.5.4/module/buffer.c <==
#include <linux/malloc.h>
#include <linux/string.h>
#include <asm/errno.h>
#include <linux/kdev_t.h>
#include <comedi_module.h>

/*#define DEBUG*/

static void dec_use(comedi_buf *buf,comedi_bufll *it);

int comedi_initbuf(comedi_buf *buf)
{
	buf->use_count=1;
	buf->in=0;
	buf->first=(comedi_bufll *)get_free_page(GFP_KERNEL);
if(!buf->first){
	printk("comedi: get_free_page failed\n");
}
	buf->last=buf->first;
	buf->num_pages=1;
	buf->first->next=NULL;
	buf->first->use_count=1;  /* 1 for the interrupt routine */

	return 0;
}

void comedi_freebuf(comedi_buf *buf)
{
	/* there *should* only be one page to free. */
	if(buf->num_pages>1){
		printk("comedi: comedi_freebuf called with %d pages in buffer\n",buf->num_pages);
	}
	if(buf->last->use_count>1){
		printk("comedi: comedi_freebuf called with last having use_count of %d\n",
			buf->last->use_count);
	}
	dec_use(buf,buf->last);
	buf->last=NULL;
}

void copy_to_buffer(comedi_buf *buf,void *src,int n)
{
	int nbytes;
	comedi_bufll *old_last;

	if(!buf->use_count)return;
	while(n>0){
		nbytes=n;
		if(nbytes>=PAGE_SIZE-sizeof(comedi_bufll)-buf->in){
			nbytes=PAGE_SIZE-sizeof(comedi_bufll)-buf->in;
			memcpy(buf->last->stuff+buf->in,src,nbytes);

			if(!(buf->last->next=
				(comedi_bufll *)get_free_page(GFP_ATOMIC)))
				return;
			buf->num_pages++;
			old_last=buf->last;
			buf->last=old_last->next;
			buf->last->use_count=buf->use_count;
#ifdef DEBUG
printk("alloc bufll 0x%08x\n",buf->last);
printk("use_count after alloc=%d\n",buf->last->use_count);
#endif
			dec_use(buf,old_last);
			buf->in=0;
		}else{
			memcpy(buf->last->stuff+buf->in,src,nbytes);
			buf->in+=nbytes;
		}
		src+=nbytes;
		n-=nbytes;
	}
}

int copy_from_buffer(comedi_bufptr *ptr,void *dest,int n)
{
	int nbytes,i,count=0;
	comedi_bufll *old_bufll;
	long flags;

	while(n>0){
		nbytes=n;
		save_flags(flags);
		cli();
		if(ptr->bufll->next==NULL){
			i=ptr->buf->in-ptr->offset;
		}else{
			i=PAGE_SIZE-sizeof(comedi_bufll)-ptr->offset;
		}
		restore_flags(flags);
		if(nbytes>i)nbytes=i;
#if OLD_KERNEL
		memcpy_tofs(dest+count,ptr->bufll->stuff+ptr->offset,nbytes);
#else
		memcpy(dest+count,ptr->bufll->stuff+ptr->offset,nbytes);
#endif
		n-=nbytes;
		count+=nbytes;
		ptr->offset+=nbytes;
		if(ptr->offset>=PAGE_SIZE-sizeof(comedi_bufll)){
			old_bufll=ptr->bufll;
			ptr->bufll=ptr->bufll->next;
			ptr->offset=0;
			dec_use(ptr->buf,old_bufll);
		}else{
			break;
		}
	}
	return count;
}

static void dec_use(comedi_buf *buf,comedi_bufll *it)
{
	it->use_count--;
#ifdef DEBUG
printk("dec_use of 0x%08x\n",it);
#endif
	if(it->use_count<=0){
		if(buf->first!=it){
			printk("comedi: freeing bufll that isn't first\n");
		}
		buf->first=it->next;
		buf->num_pages--;
		free_page((long)it);
	}
}

comedi_bufptr *comedi_hookbuf(comedi_buf *buf)
{
	comedi_bufptr *it;

	it=(comedi_bufptr *)kmalloc(sizeof(comedi_bufptr),GFP_KERNEL);
	it->buf=buf;
	buf->use_count++;

	it->bufll=buf->last;  /* these 3 lines should be atomic */
	it->offset=buf->in;
	it->bufll->use_count++;
#ifdef DEBUG
printk("inc_use of 0x%08x\n",it->bufll);
printk("it->bufll->use_count after hookup=%d\n",it->bufll->use_count);
#endif

	return it;
}

void comedi_unhookbuf(comedi_bufptr *it)
{
	comedi_bufll *next,*tmp;

	it->offset=0;
	next=it->bufll;
	while(next){
		tmp=next->next;
		dec_use(it->buf,next);
		next=tmp;
	}
	it->buf->use_count--;
	kfree(it);
}

==> comedi-0.6.2.tgz: comedi-0.6.2/module/buffer.c <==
#include <linux/malloc.h>
#include <linux/string.h>
#include <asm/errno.h>
#include <linux/kdev_t.h>
#include <asm/uaccess.h>
#include <comedi_module.h>

/*#define DEBUG*/

static void dec_use(comedi_buf *buf,comedi_bufll *it);

int comedi_initbuf(comedi_buf *buf)
{
	buf->use_count=1;
	buf->in=0;
	buf->first=(comedi_bufll *)get_free_page(GFP_KERNEL);
if(!buf->first){
	printk("comedi: get_free_page failed\n");
}
	buf->last=buf->first;
	buf->num_pages=1;
	buf->first->next=NULL;
	buf->first->use_count=1;  /* 1 for the interrupt routine */

	return 0;
}

void comedi_freebuf(comedi_buf *buf)
{
	/* there *should* only be one page to free. */
	if(buf->num_pages>1){
		printk("comedi: comedi_freebuf called with %d pages in buffer\n",buf->num_pages);
	}
	if(buf->last->use_count>1){
		printk("comedi: comedi_freebuf called with last having use_count of %d\n",
			buf->last->use_count);
	}
	dec_use(buf,buf->last);
	buf->last=NULL;
}

void copy_to_buffer(comedi_buf *buf,void *src,int n)
{
	int nbytes;
	comedi_bufll *old_last;

	if(!buf->use_count)return;
	while(n>0){
		nbytes=n;
		if(nbytes>=PAGE_SIZE-sizeof(comedi_bufll)-buf->in){
			nbytes=PAGE_SIZE-sizeof(comedi_bufll)-buf->in;
			memcpy(buf->last->stuff+buf->in,src,nbytes);

			if(!(buf->last->next=
				(comedi_bufll *)get_free_page(GFP_ATOMIC)))
				return;
			buf->num_pages++;
			old_last=buf->last;
			buf->last=old_last->next;
			buf->last->use_count=buf->use_count;
#ifdef DEBUG
printk("alloc bufll 0x%08x\n",buf->last);
printk("use_count after alloc=%d\n",buf->last->use_count);
#endif
			dec_use(buf,old_last);
			buf->in=0;
		}else{
			memcpy(buf->last->stuff+buf->in,src,nbytes);
			buf->in+=nbytes;
		}
		src+=nbytes;
		n-=nbytes;
	}
}

int copy_from_buffer(comedi_bufptr *ptr,void *dest,int n)
{
	int nbytes,i,count=0;
	comedi_bufll *old_bufll;
	long flags;

	while(n>0){
		nbytes=n;
		save_flags(flags);
		cli();
		if(ptr->bufll->next==NULL){
			i=ptr->buf->in-ptr->offset;
		}else{
			i=PAGE_SIZE-sizeof(comedi_bufll)-ptr->offset;
		}
		restore_flags(flags);
		if(nbytes>i)nbytes=i;
#if KERNEL_V20
		memcpy_tofs(dest+count,ptr->bufll->stuff+ptr->offset,nbytes);
#else
		copy_to_user(dest+count,ptr->bufll->stuff+ptr->offset,nbytes);
#endif
		n-=nbytes;
		count+=nbytes;
		ptr->offset+=nbytes;
		if(ptr->offset>=PAGE_SIZE-sizeof(comedi_bufll)){
			old_bufll=ptr->bufll;
			ptr->bufll=ptr->bufll->next;
			ptr->offset=0;
			dec_use(ptr->buf,old_bufll);
		}else{
			break;
		}
	}
	return count;
}

static void dec_use(comedi_buf *buf,comedi_bufll *it)
{
	it->use_count--;
#ifdef DEBUG
printk("dec_use of 0x%08x\n",it);
#endif
	if(it->use_count<=0){
		if(buf->first!=it){
			printk("comedi: freeing bufll that isn't first\n");
		}
		buf->first=it->next;
		buf->num_pages--;
		free_page((long)it);
	}
}

comedi_bufptr *comedi_hookbuf(comedi_buf *buf)
{
	comedi_bufptr *it;

	it=(comedi_bufptr *)kmalloc(sizeof(comedi_bufptr),GFP_KERNEL);
	it->buf=buf;
	buf->use_count++;

	it->bufll=buf->last;  /* these 3 lines should be atomic */
	it->offset=buf->in;
	it->bufll->use_count++;
#ifdef DEBUG
printk("inc_use of 0x%08x\n",it->bufll);
printk("it->bufll->use_count after hookup=%d\n",it->bufll->use_count);
#endif

	return it;
}

void comedi_unhookbuf(comedi_bufptr *it)
{
	comedi_bufll *next,*tmp;

	it->offset=0;
	next=it->bufll;
	while(next){
		tmp=next->next;
		dec_use(it->buf,next);
		next=tmp;
	}
	it->buf->use_count--;
	kfree(it);
}
